/*
 * Ngmc.cpp
 *
 *  Created on: Feb 17, 2011
 *      Author: philipp_
 */
#include "Ngmc.h"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string.h>

#include "datatypes.h"
#include "FastaParser.h"
#include "FastqParser.h"
#include "NgmRefWriter.h"
#include "NgmReadWriter.h"

#include <map>
#include <string>

using std::map;
using std::string;

using std::cout;
using std::endl;

bool const debug = false;

Parser * getParser(char *seqfile) {
	Parser * parser = 0;
	string line;
	ifstream seqFile(seqfile);
	if (seqFile.is_open() && seqFile.good()) {
		getline(seqFile, line);
		if (line.length() > 0) {
			switch (line[0]) {
				case '>':
					parser = new FastaParser();
					break;
				case '@':
					parser = new FastqParser();
					break;
				default:
					throw "Unable to detect input format.\nSupported formats: fasta, fastq.";
			}
		} else {
			throw "Unable to detect input format.\nSupported formats: fasta, fastq.";
		}
		seqFile.close();
	} else {
		throw string("Couldn't open file ") + seqfile;
	}
	parser->init();
	return parser;
}

bool isNucleotide(char const c) {
	return c == 'A' || c == 'a' || c == 'C' || c == 'c' || c == 'G' || c == 'g' || c == 'T' || c == 't' || c == 'N' || c == 'n';
}

string parsePairId(string id) {
	return id.substr(0, id.rfind('/'));
}

void cleareEntry(Entry * & entry) {
	if (entry != 0) {
		if (entry->sequence != 0) {
			delete entry->sequence;
			entry->sequence = 0;
		}
		delete entry;
		entry = 0;
	}
}

bool checkRead(Entry const * const entry) {
	bool valid = entry->sequenceLength > 0 && entry->nameLength > 0;
	if (valid) {
		for (int i = 0; i < entry->sequenceLength && valid; ++i) {
			valid = valid && isNucleotide(entry->sequence[i]);
		}
	}
	return valid;
}

unsigned int computeCheckSum(unsigned long const fileLength1, unsigned long const fileLength2, int const readCount) {
	return (fileLength1 + fileLength2) % readCount;
}

int convertRef(Parser * parser, NgmRefWriter & writer, int padding, int readCount) {

	int processedRead = 0;
	int flags = 0;
	writer.writeRefHeader(readCount, flags, padding);

	Entry * entry = new Entry;
	int readsWritten = 0;
	string name;
	while (parser->read(name, entry)) {
		if (processedRead / 10000.0 == 0) {
			//cout << "Progress: " << (((processedRead++ * 100.0 / readCount) * 100) / 100.0) << "\r";
		}
		bool valid = checkRead(entry);
		FilePosition outputPosition = writer.writeReference(entry->sequence, entry->sequenceLength, padding);
		writer.writeIndexEntryRef(outputPosition + padding, entry->sequenceLength, entry->nameOffset, entry->nameLength, valid, valid);
		readsWritten += 1;
		cleareEntry(entry);
		entry = new Entry;
	}
	//cout << endl << endl;

	return readsWritten;
}

int convert(Parser * parser, NgmReadWriter & writer, int maxReadLength, int readCount, vector<Statistics*> &jops) {
	string name;
	int readsWritten = 0;
	int processedRead = 0;

	int flags = 0;
	writer.writeReadHeader(maxReadLength, readCount, flags);

	Entry * entry = new Entry;
	while (parser->read(name, entry)) {
		for (vector<Statistics*>::iterator i = jops.begin(); i != jops.end(); i++) {
			(*(*i)).add(entry->sequence);
		}
		bool valid = checkRead(entry);
		if (!valid) {
			if (strlen(entry->sequence) == 0) {
				std::cerr << "Invalid read found: empty read" << endl;
			} else {
				std::cerr << "Invalid read found (" << strlen(entry->sequence) << "bp): " << entry->sequence << endl;
			}

		}
		writer.writeRead(entry->sequence, maxReadLength);
		writer.writeIndexEntryRead(entry, valid);
		cleareEntry(entry);
		entry = new Entry;
		if (valid)
			readsWritten += 1;

		if (processedRead / 10000.0 == 0) {
			//cout << "Progress: " << (((processedRead++ * 100.0 / readCount) * 100) / 100.0) << "\r";
		}
	}
	//cout << endl << endl;

	return readsWritten;
}

bool parseNext(map<string, Entry *> & pairs, Parser * parser, int const maxReadLength, NgmReadWriter & writerLeft, NgmReadWriter & writerRight, bool isFirst, int & writtenReads, vector<Statistics*> & jops_left, vector<Statistics*> & jops_right) {
	Entry * entryLeft = new Entry;
	entryLeft->first = isFirst;
	string nameLeft;

	if (parser->read(nameLeft, entryLeft)) {
		Entry * entryRight = 0;
		nameLeft = parsePairId(nameLeft);
		if (pairs.find(nameLeft) != pairs.end()) {
			if (entryLeft->first) {
				entryRight = pairs[nameLeft];
				if (pairs[nameLeft]->first) {
					cout << "Warning: Found read pair in a single file (Name: " << nameLeft << ")! Do you have two reads with the same name in one file?" << std::endl;
				}
			} else {
				entryRight = entryLeft;
				entryLeft = pairs[nameLeft];
				if (!pairs[nameLeft]->first) {
					cout << "Warning: Found read pair in a single file (Name: " << nameLeft << ")! Do you have two reads with the same name in one file?" << std::endl;
				}
			}

			FilePosition outputPositionLeft = writerLeft.writeRead(entryLeft->sequence, maxReadLength);
			//left
			for (vector<Statistics*>::iterator i = jops_left.begin(); i != jops_left.end(); i++) {
				(*(*i)).add(entryLeft->sequence);
			}

			//TODO: Ask if here should really be entryLeft???
			FilePosition outputPositionRight = writerRight.writeRead(entryRight->sequence, maxReadLength);
			//right
			for (vector<Statistics*>::iterator i = jops_right.begin(); i != jops_right.end(); i++) {
				(*(*i)).add(entryRight->sequence);
			}
			writtenReads += 1;
			if (outputPositionLeft != outputPositionRight) {
				//TODO:
				throw "Invalid output position in function parseNext.";
			}
			writerLeft.writeIndexEntryRead(entryLeft, checkRead(entryLeft));
			writerRight.writeIndexEntryRead(entryRight, checkRead(entryRight));

			pairs.erase(nameLeft);
			cleareEntry(entryLeft);
			cleareEntry(entryRight);
		} else {
			pairs[nameLeft] = entryLeft;
		}
		return true;
	} else {
		cleareEntry(entryLeft);
		return false;
	}
}

int convertPairs(Parser * parser, Parser * parserRight, NgmReadWriter & writerLeft, NgmReadWriter & writerRight, int maxReadLengthLeft, int maxReadLengthRight, unsigned int const checksum, vector<Statistics*> & jops_left, vector<Statistics*> & jops_right) {

	int flags = 0x00000001;
	writerLeft.writeReadHeader(maxReadLengthLeft, 0, flags, checksum);
	writerRight.writeReadHeader(maxReadLengthRight, 0, flags, checksum);

	map<string, Entry *> pairs;

	int pairNumber = 0;
	bool eof = false;
	while (!eof) {
		bool foundLeft = parseNext(pairs, parser, maxReadLengthLeft, writerLeft, writerRight, true, pairNumber, jops_left, jops_right);
		bool foundRigh = parseNext(pairs, parserRight, maxReadLengthRight, writerLeft, writerRight, false, pairNumber, jops_left, jops_right);

		eof = !foundLeft && !foundRigh;
	}
	int writtenReads = 2 * pairNumber + pairs.size();
	int lineNumber = pairNumber + pairs.size();
	if (pairs.size() > 0) {
		for (map<string, Entry *>::iterator it = pairs.begin(); it != pairs.end(); it++) {
			Entry * entry = it->second;

			bool valid = checkRead(entry);
			if (entry->first) {
				writerLeft.writeIndexEntryRead(entry, valid);
				writerLeft.writeRead(entry->sequence, maxReadLengthLeft);
				//left
				for (vector<Statistics*>::iterator i = jops_left.begin(); i != jops_left.end(); i++) {
					(*(*i)).add(entry->sequence);
				}
				writerRight.writeDummy(maxReadLengthLeft);
				//Dummy entry: has to be ignored
				writerRight.writeIndexEntryRead(entry, false);
			} else {
				writerRight.writeRead(entry->sequence, maxReadLengthRight);
				writerRight.writeIndexEntryRead(entry, valid);
				//right
				for (vector<Statistics*>::iterator i = jops_right.begin(); i != jops_right.end(); i++) {
					(*(*i)).add(entry->sequence);
				}
				writerLeft.writeDummy(maxReadLengthRight);
				//Dummy entry: has to be ignored
				writerLeft.writeIndexEntryRead(entry, false);
			}
			//			if (outputPosition != outputPositionDummy) {
			//				//TODO:
			//				throw "Invalid output position in function convertPairs.";
			//			}
		}
	}

	//Write header again with correct read number
	writerLeft.reWriteReadHeader(maxReadLengthLeft, lineNumber, flags, checksum);
	writerRight.reWriteReadHeader(maxReadLengthRight, lineNumber, flags, checksum);

	cout << "Found " << pairs.size() << " reads without a mate." << std::endl;
	return writtenReads;
}
