#ifndef __CS_H__
#define __CS_H__

#include "NGM.h"

#include "SequenceLocation.h"
#include "MappedRead.h"
#include "RefEntry.h"

#include <map>
#include <stdlib.h>
#include <string>

#include "CSCache.h"

/*
Candidate Search:

Initialisierung:
  Iteriere ueber jeden Praefix der definierten Referenzen (config:ref_mode mit -1=alle, 0..n eine einzelne referenz)
    Fuege in RefTable fuer diesen Praefix seine Position und die Id des Referenzstrings ein

Suche:
  Iteriere ueber jeden Pruefix des Reads
    Lies fuer diesen Praefix die Positionsliste aus RefTable
	Fuer jede Position dieses Praefixes
	  Offset um die Position des Praefixes (=normalisierung auf den Read-Anfang)
	  fast_mode:
	    Pruefe rTable-Capactiy, wenn overflow -> Neustart der Suche im safe_mode
	    Schreibe in rTable mit linearem Kollisionshandling
	  save_mode:
	    Schreibe in iTable

  Sammle Ergebnisse ein:
    Iteriere ueber jedes Element in fast_mode:rTable / save_mode:iTable
	  Wenn count > threshhold, schicke es zur weiterverarbeitung

*/
class CS : public NGMTask
{
private:
	volatile static int s_ThreadCount;
	int const m_CSThreadID;
	int const m_BatchSize;
	std::vector<MappedRead*> m_CurrentBatch;
	int m_CurrentSeq;
	int m_CurrentPage;
	int m_ProcessedReads;
	int m_WrittenReads;
	int m_DiscardedReads;
	volatile int m_Candidates;
	int m_Mode;
	bool m_EnableBS;
	int m_UsedSlots;
	int m_TotalLocs;
	int m_TotalPrefixes;
	int m_Overflows;

	IRefProvider const * m_RefProvider;

	int const m_SrchTableLen;
	int const m_Threshhold;
	int const m_REThresh;
	uint const m_PrefixBaseSkip;
	bool m_Fallback;

	static uint prefixBasecount;
	static uint prefixBits;
	static ulong prefixMask;

	typedef void (*PrefixIterationFn)(ulong prefix, int pos, void* data);
	typedef void (CS::*AddLocationFn)(SequenceLocation const & loc);

	static void BuildPrefixTable(ulong prefix, int pos, void* data);
	static void PrefixSearch(ulong prefix, int pos, void* data);
	static void PrefixMutateSearch(ulong prefix, int pos, void* data);
	static void PrefixMutateSearchEx(ulong prefix, int pos, void* data, int mpos = 0);
	AddLocationFn AddLocation;
	void AddLocationStd(SequenceLocation const & loc);
	void AddLocationFallback(SequenceLocation const & loc);
	int CollectResultsStd(MappedRead * read);
	int CollectResultsFallback(MappedRead * read);
	void FilterScore(LocationScore * score);

	void CheckFallback();

	void RunBatch();
	void SendToBuffer(MappedRead * read);


	void CheckRTable(); // Debug/Validation

	LocationScore * rTable;	// standard
	std::map<SequenceLocation, int> iTable;		// fallback

	inline int Hash(int n);

	static inline int calc_binshift(int corridor)
	{
		corridor >>= 1;
		int l = 0;
		while ( (corridor >>= 1) > 0 )
			++l;
		return l;
	}

	inline int GetBin(int pos)
	{
		static int shift = calc_binshift(Config.GetInt("corridor"));
		return pos >> shift;
	}
	inline int ResolveBin(int bin)
	{
		static int shift = calc_binshift(Config.GetInt("corridor"));
		static int offset = (shift > 0) ? 1 << (shift - 1) : 0;
		return (bin << shift) + offset;
	}
public:
	inline const int GetMode() { return m_Mode; }
	inline static const uint GetPrefixLength() { return prefixBasecount; }

	static void Init();
	static void Cleanup();
	static void PrefixIteration(char const * sequence, uint length, PrefixIterationFn func, void* data, uint prefixskip = 0, uint offset = 0);

	CS(bool useBuffer = true);
	~CS();

	void DoRun();
	inline int GetStage() const { return 0; }
	inline char const * GetName() const { return "CS"; }
};

/*static inline int CSHash(ulong prefix)
{
	return prefix % CS::cRefTableLen;
}*/

static const int cInvalidLocation = -99999999;
static const SequenceLocation sInvalidLocation = { cInvalidLocation, 0 };
static const LocationScore sInvalidLocationScore (sInvalidLocation, 0, 0); // { sInvalidLocation, {0} };

#endif
