#include <stdio.h>
#include <memory.h>

#include "RefEntry.h"
#include "Types.h"

#include "Log.h"
#include "NGM.h"

#include "PrefixTable.h"

#include "CS.h"
#include "CSCache.h"

#undef module_name
#define module_name "CSCache"

const uint cMaxCachePages = 62;

struct CacheIndex
{
	uint PrefixLength;
	uint RefTableLength;
	uint CheckSum;
	uint PageCount;
	ulong Reserved;
	ulong MaxPageSize;	// To allocate memory when loading a page
};
struct PageIndex
{
	ulong Offset;
	ulong Length;
};

CSCache * CSCache::CreateCache(std::string const filename, uint const pages)
{
	CSCache * cache = new CSCache(filename);
	cache->m_FP = fopen(filename.c_str(), "w+b");

	if (cache->m_FP == 0)
	{
		Log.Error("Unable to create page cache file.");
		Fatal();
	}

	cache->m_Index = new CacheIndex();
	size_t const index_size = sizeof(CacheIndex);
	memset(cache->m_Index, 0, index_size);

	cache->m_Index->PrefixLength = CS::GetPrefixLength();
	cache->m_Index->RefTableLength = PrefixTable::cRefTableLen;
	cache->m_Index->PageCount = pages;
	cache->m_Index->MaxPageSize = 0;

	cache->m_PageIndex = new PageIndex[pages];
	memset(cache->m_PageIndex, 0, pages * sizeof(PageIndex));

	cache->CommitHeader();

	return cache;
}

CSCache * CSCache::LoadCache(std::string const filename)
{
	CSCache * cache = new CSCache(filename);
	cache->m_FP = fopen(filename.c_str(), "rb");

	if (cache->m_FP == 0)
	{
		Log.Error("Unable to open page cache file.");
		return 0;
	}

	size_t const index_size = sizeof(CacheIndex);
	char * buf = new char[index_size];

	size_t read = fread(buf, 1, index_size, cache->m_FP);
	if (read != index_size)
	{
		Log.Error("Error reading cache file header");
		return 0;
	}

	cache->m_Index = (CacheIndex*)buf;

	if (cache->m_Index->PrefixLength != CS::GetPrefixLength())
	{
		Log.Error("Mismatching prefix size in cache (Cache: %i Bases, Config: %i Bases)", cache->m_Index->PrefixLength, CS::GetPrefixLength());
		Fatal();
	}

	size_t const pageindex_size = sizeof(PageIndex) * cache->m_Index->PageCount;
	buf = new char[pageindex_size];
	if (pageindex_size != fread(buf, 1, pageindex_size, cache->m_FP))
	{
		Log.Error("Error reading page index");
		return 0;
	}
	cache->m_PageIndex = (PageIndex*)buf;

	return cache;
}

void CSCache::CommitHeader()
{
	rewind(m_FP);

	if (fwrite(m_Index, sizeof(CacheIndex), 1, m_FP) != 1)
	{
		Log.Error("Error writing cache file header.");
		Fatal();
	}

	if (fwrite(m_PageIndex, m_Index->PageCount * sizeof(PageIndex), 1, m_FP) != 1)
	{
		Log.Error("Error writing page index.");
		Fatal();
	}
	fflush(m_FP);
}

void CSCache::AddPage(int const page, char const * pBuffer, ulong const length)
{
	if ((uint)page > m_Index->PageCount)
	{
		Log.Error("Invalid page %i (out of range)", page);
		return;
	}

	if (m_PageIndex[page].Offset != 0)
	{
		Log.Error("Page %i already in use", page);
		return;
	}

	Log.Verbose("Adding cache page %i", page);
	NGMLock(&m_Mutex);
	fseek(m_FP, 0, SEEK_END);

	ulong pageStart = ftell(m_FP);

	if (fwrite(pBuffer, 1, length, m_FP) != length)
	{
		Log.Error("Error writing page %i to file", page);
		NGMUnlock(&m_Mutex);
		return;
	}

	if (length > m_Index->MaxPageSize)
		m_Index->MaxPageSize = length;

	m_PageIndex[page].Offset = pageStart;
	m_PageIndex[page].Length = length;

	CommitHeader();
	NGMUnlock(&m_Mutex);
}

void CSCache::LoadPage(int const page, char const * & pBuffer, ulong & length) const
{
	if (!ContainsData(page))
	{
		Log.Error("Page contains no data");
		return;
	}

	ulong const page_size = m_PageIndex[page].Length;

	char * buf = new char[page_size];
	fseek(m_FP, m_PageIndex[page].Offset, SEEK_SET);
	if (fread(buf, 1, page_size, m_FP) != page_size)
	{
		Log.Error("Error reading page %i (Offset %lu, Length %lu)", page, m_PageIndex[page].Offset, page_size);
	}

	pBuffer = buf;
	length = page_size;
}

int CSCache::GetPageCount() const
{
	return m_Index->PageCount;
}

bool CSCache::ContainsData(int const page) const
{
	return ((m_PageIndex[page].Offset != 0) && (m_PageIndex[page].Length != 0));
}

CSCache::CSCache(std::string const filename) :
		m_Filename( filename ),
		m_Index( 0 ),
		m_PageIndex( 0 ),
		m_Mutex()
{
	NGMInitMutex(&m_Mutex);
}
CSCache::~CSCache()
{
	fclose(m_FP);
	if (m_Index != 0)
		delete m_Index;

	if (m_PageIndex != 0)
		delete[] m_PageIndex;
}
