#include "MappedRead.h"
#include "Log.h"

#undef module_name
#define module_name "READ"

#ifdef INSTANCE_COUNTING
volatile int MappedRead::sInstanceCount = 0;
volatile int LocationScore::sInstanceCount = 0;
#endif

MappedRead::MappedRead(int const readid) :
	ReadId(readid),
	Calculated( 0 ),
	TopScore( -1 ),
	Lock( 0 ),
	EqualScoringID( 0 ),
	EqualScoringCount( 1 ),
	Scores(),
	Paired( 0 ),
	Identity( 0 ),
	Strand( '?' ),
	QStart( 0 ),
	QEnd( 0 ),
	Seq( 0 ),
	Buffer1( 0 ),
	Buffer2( 0 )
{
#ifdef INSTANCE_COUNTING
		AtomicInc(&sInstanceCount);
#endif
}

void MappedRead::AllocBuffers()
{
	Buffer1 = new char[std::max(1, SequenceProvider.GetReadMaxLen())*4];
	Buffer2 = new char[std::max(1, SequenceProvider.GetReadMaxLen())*4];
	*(int*)Buffer1 = 0x212121;
	*(int*)Buffer2 = 0x212121;
}

MappedRead::~MappedRead()
{
	for (size_t i = 0; i < Scores.size(); ++i)
		delete Scores[i];
	delete[] Buffer1;
	delete[] Buffer2;
#ifdef INSTANCE_COUNTING
	AtomicDec(&sInstanceCount);
#endif
}

LocationScore * MappedRead::AddScore(LocationScore const & score)
{
	LocationScore * toInsert = new LocationScore(score);
	toInsert->Read = this;
	Scores.push_back(toInsert);
	return toInsert;
}

LocationScore * MappedRead::TLS() const
{
	if (TopScore == -1)
	{
		Log.Error("Top score accessed without calculation (Read %i, ESID %i, ESC %i)", ReadId, EqualScoringID, EqualScoringCount);
		return 0;
	}
	return Scores[TopScore];
}

int MappedRead::nScores() const
{
	return Scores.size();
}

void PrintScores(MappedRead const * read)
{
	Log.Green("Scores for read %i", read->ReadId);
	for (size_t i = 0; i < read->Scores.size(); ++i)
	{
		if (read->Scores[i] == 0)
			Log.Message("<0>");
		else
			Log.Message("Score %.2f for <%i, %i>", read->Scores[i]->Score.f, read->Scores[i]->Location.m_Location, read->Scores[i]->Location.m_RefId);
	}
}

void MappedRead::TopN(uint n)
{
	if (Scores.size() == 0)
		return;

	if (Calculated > 0 && Calculated != (int)Scores.size())
		Log.Error("TopN called for inconsistent scores may lead to unexpected results.");

	std::sort(Scores.begin(), Scores.end(), SortPred); // Sort scores from best to worst
	Unique(); // Remove double scores
	Scores.erase(std::remove_if(Scores.begin(), Scores.end(), IsZero), Scores.end()); // remove 0-pointer from last step

	// remove from back all but top n
	while (Scores.size() > n)
	{
		delete Scores[Scores.size() - 1];
		Scores.pop_back();
	}

	// realloc vector
	std::vector<LocationScore*>(Scores).swap(Scores);

	TopScore = 0;
	Calculated = n;
}

void MappedRead::Unique()
{
	uint ci = 0;
	for (uint i = 1; i < Scores.size(); ++i)
	{
		if (UniquePred(Scores[ci], Scores[i]))
		{
			delete Scores[i]; Scores[i] = 0;
		}
		else
			ci = i;
	}
}
