#include "Types.h"
#include "RefEntry.h"
#include "PrefixTable.h"
#include "Log.h"
#include "CS.h"

#include <memory.h>

#undef module_name
#define module_name "PFLT"

struct TableHeader
{
	uint Magic;
	uint Checksum;
	uint PrefixSize;
	uint HashTableLen;
	uint MaxPrefixLocs;
	uint TotalPrefixCount;
	ulong TotalSize;
	ulong TotalLocCount;
};

struct PrefixEntry
{
	ulong Prefix;
	uint Count;
	uint Next;
};

uint const cSigMore = 0xF11FF11F;
uint const cSigLast = 0xF00FF00F;
uint const cSigStart = 0x22FFFF11;
uint const cSigEnd = 0x44FFFF33;

ulong PrefixTable::CalcSize() const
{
	return sizeof(TableHeader) +
			PrefixTable::cRefTableLen * sizeof(uint) +
			this->m_RRCount * sizeof(PrefixEntry) +
			this->m_TotalLocs * sizeof(SequenceLocation) +
			sizeof(uint);
}
int FlattenPrefix(char * buffer, PrefixRefRelay * prefix);

char const * PrefixTable::Flatten(ulong & resultSize) const
{
	ulong bsize = CalcSize();
	char * buf = new char[bsize];
	memset(buf, 0, bsize);

	return Flatten(buf, bsize, resultSize);
}


char const * PrefixTable::Flatten(char * pBuffer, ulong const bufferSize, ulong & resultSize) const
{
	ulong calcSize = CalcSize();

	if (calcSize > bufferSize)
	{
		Log.Error("Prefix Table flattening failed: Buffer too small");
		return 0;
	}

	TableHeader * header = (TableHeader*)pBuffer;
	uint curPos = sizeof(TableHeader);

	header->Magic = cSigStart;
	header->Checksum = 0;
	header->PrefixSize = CS::GetPrefixLength();
	header->HashTableLen = PrefixTable::cRefTableLen;

	uint * indexTable = (uint*)(pBuffer + curPos);
	curPos += PrefixTable::cRefTableLen * sizeof(uint);

	for (uint i = 0; i < PrefixTable::cRefTableLen; ++i)
	{
		PrefixRefRelay * cur = RefTable[i];

		if (cur != 0)
		{
			indexTable[i] = curPos;
			while (cur != 0)
			{
				curPos += FlattenPrefix(pBuffer+curPos, cur);
				++header->TotalPrefixCount;
				if (header->MaxPrefixLocs < (uint)cur->Entry->refTotal)
					header->MaxPrefixLocs = cur->Entry->refTotal;
				header->TotalLocCount += cur->Entry->refTotal;
				cur = cur->Next;
			}
		}
	}
	*((int*)(pBuffer+curPos)) = cSigEnd;
	curPos += 4;

	header->TotalSize = resultSize = curPos;

	if (curPos != calcSize)
		Log.Error("Size mismatch (Calculated %lu, Got %lu)", calcSize, curPos);

	return pBuffer;
}

int FlattenPrefix(char * buffer, PrefixRefRelay * prefix)
{
	uint written = 0;
	PrefixEntry * entry = (PrefixEntry*)buffer;
	buffer += sizeof(PrefixEntry);
	written = sizeof(PrefixEntry);

	entry->Prefix = prefix->Prefix;
	entry->Next = (prefix->Next == 0) ? cSigLast : cSigMore;
	entry->Count = prefix->Entry->refTotal;

	//Log.Verbose("Writing prefix 0x%x (%i locs)", prefix->Prefix, entry->Count);

	RefEntry * cur = prefix->Entry;

	uint elementsWritten = 0;
	while (cur != 0)
	{
		int sz = sizeof(SequenceLocation) * cur->refCount;
		memcpy(buffer, cur->ref, sz);
		elementsWritten += cur->refCount;
		buffer += sz;
		written +=sz;
		cur = cur->nextEntry;
	}

	if (elementsWritten != entry->Count)
		Log.Error("Mismatching write count on prefix 0x%x (wrote %i, expected %i)", prefix->Prefix, elementsWritten, entry->Count);

	if (written != (sizeof(PrefixEntry) + sizeof(SequenceLocation) * prefix->Entry->refTotal))
	{
		Log.Error("Inconsistent write for prefix 0x%x (%i locs)", prefix->Prefix, entry->Count);
	}

	return written;
}

PrefixTable::PrefixTable() : m_RECount(0), m_RRCount(0), m_BCalls(0)
{
	RefTable = new PrefixRefRelay*[cRefTableLen];
	memset(RefTable, 0, sizeof(PrefixRefRelay*)*cRefTableLen);
}

RefEntry::RefEntry(void const * pEntries, int count)
{
	refCount = refTotal = count;
	nextEntry = 0;
	ref = new SequenceLocation[count];
	memcpy(ref, pEntries, count*sizeof(SequenceLocation));
}

bool PrefixTable::CheckHeader(char const * const buffer)
{
	TableHeader * header = (TableHeader*)buffer;

	if (header->Magic != cSigStart)
	{
		Log.Error("Invalid header signature");
		return false;
	}

	if (header->HashTableLen != cRefTableLen)
	{
		Log.Error("Mismatching ref table lengths (Cache: %lu, Expected: %lu)", header->HashTableLen, cRefTableLen);
		return false;
	}
	if (header->PrefixSize != CS::GetPrefixLength())
	{
		Log.Error("Mismatching prefix size (Cache: %u, Expected: %u)", header->PrefixSize, CS::GetPrefixLength());
		return false;
	}

	return true;
}

void PrefixTable::SaveToCache(CSCache * cache, int const page)
{
	ulong bufferSize = 0;
	Log.Message("Flattening %i", page);
	char const * buffer = Flatten(bufferSize);

	Log.Message("Writing %i", page);
	cache->AddPage(page, buffer, bufferSize);

	delete[] buffer;
}

PrefixTable * PrefixTable::GenerateFromCache(CSCache const * cache, int const page)
{
	ulong blen = 0;
	char const * pBuffer = 0;
	cache->LoadPage(page, pBuffer, blen);

	PrefixTable * table = GenerateFromBuffer(pBuffer, blen);

	delete[] pBuffer;

	return table;
}

PrefixTable * PrefixTable::GenerateFromBuffer(char const * pBuffer, ulong const length)
{
	bool checkHashes = Config.Exists("checkhashes") && (Config.GetInt("checkhashes") != 0);

	TableHeader * header = (TableHeader*)pBuffer;
	uint * indexTable = (uint*)(pBuffer + sizeof(TableHeader));

	if (!CheckHeader(pBuffer))
		Fatal();

	PrefixTable * table = new PrefixTable();
	for (uint i = 0; i < header->HashTableLen; ++i)
	{
		ulong offset = indexTable[i];

		while (offset != 0)
		{
			PrefixEntry * entry = (PrefixEntry*)(pBuffer + offset);
			offset += sizeof(PrefixEntry);

			if (checkHashes && (Hash(entry->Prefix) != i))
				Log.Error("Hash mismatch for prefix 0x%x", entry->Prefix);

			//Log.Verbose("Loading prefix 0x%x (count = %i)", entry->Prefix, entry->Count);

			PrefixRefRelay * prefix = new PrefixRefRelay();
			prefix->Prefix = entry->Prefix;
			prefix->Entry = new RefEntry(pBuffer+offset, entry->Count);

			prefix->Next = table->RefTable[i];
			table->RefTable[i] = prefix;
			++table->m_RRCount;
			++table->m_RECount;

			if (entry->Next == cSigMore)
			{
				offset += sizeof(SequenceLocation) * entry->Count;
			}
			else
			{
				offset = 0;
			}
		}
	}

	Log.Green("Prefix table loaded from cache (%i Prefixes, %i Locs)", table->m_RRCount, table->m_RECount);

	return table;
}
